@@ -4,19 +4,16 @@ kombu.messaging
 
 Sending and receiving messages.
 
-:copyright: (c) 2009 - 2012 by Ask Solem.
-:license: BSD, see LICENSE for more details.
-
 """
 from __future__ import absolute_import
 
 from itertools import count
 
-from .connection import maybe_channel
-from .entity import Exchange, Queue
+from .connection import maybe_channel, is_connection
+from .entity import Exchange, Queue, DELIVERY_MODES
 from .compression import compress
 from .serialization import encode
-from .utils import maybe_list
+from .utils import ChannelPromise, maybe_list
 
 __all__ = ['Exchange', 'Queue', 'Producer', 'Consumer']
 
@@ -42,13 +39,11 @@ class Producer(object):
         Note that the producer needs to drain events to use this feature.
 
     """
-    #: The connection channel used.
-    channel = None
 
-    #: Default exchange.
+    #: Default exchange
     exchange = None
 
-    # Default routing key.
+    #: Default routing key.
     routing_key = ''
 
     #: Default serializer to use. Default is JSON.
@@ -65,22 +60,27 @@ class Producer(object):
     #: Basic return callback.
     on_return = None
 
+    #: Set if channel argument was a Connection instance (using
+    #: default_channel).
+    __connection__ = None
+
     def __init__(self, channel, exchange=None, routing_key=None,
             serializer=None, auto_declare=None, compression=None,
             on_return=None):
-        self.channel = channel
-        self.exchange = exchange or self.exchange
-        if self.exchange is None:
-            self.exchange = Exchange('')
+        self._channel = channel
+        self.exchange = exchange
         self.routing_key = routing_key or self.routing_key
         self.serializer = serializer or self.serializer
         self.compression = compression or self.compression
         self.on_return = on_return or self.on_return
+        self._channel_promise = None
+        if self.exchange is None:
+            self.exchange = Exchange('')
         if auto_declare is not None:
             self.auto_declare = auto_declare
 
-        if self.channel:
-            self.revive(self.channel)
+        if self._channel:
+            self.revive(self._channel)
 
     def __reduce__(self):
         return self.__class__, self.__reduce_args__()
@@ -133,7 +133,7 @@ class Producer(object):
         :keyword retry: Retry publishing, or declaring entities if the
             connection is lost.
         :keyword retry_policy: Retry configuration, this is the keywords
-            supported by :meth:`~kombu.connection.Connection.ensure`.
+            supported by :meth:`~kombu.Connection.ensure`.
         :keyword \*\*properties: Additional message properties, see AMQP spec.
 
         """
@@ -141,44 +141,76 @@ class Producer(object):
         retry_policy = {} if retry_policy is None else retry_policy
         routing_key = self.routing_key if routing_key is None else routing_key
         compression = self.compression if compression is None else compression
+        exchange = exchange or self.exchange
 
         if isinstance(exchange, Exchange):
+            delivery_mode = delivery_mode or exchange.delivery_mode
             exchange = exchange.name
+        else:
+            delivery_mode = delivery_mode or self.exchange.delivery_mode
+        if not isinstance(delivery_mode, (int, long)):
+            delivery_mode = DELIVERY_MODES[delivery_mode]
+        properties['delivery_mode'] = delivery_mode
 
         body, content_type, content_encoding = self._prepare(
                 body, serializer, content_type, content_encoding,
                 compression, headers)
-        message = self.exchange.Message(body,
-                                        delivery_mode,
-                                        priority,
-                                        content_type,
-                                        content_encoding,
-                                        headers=headers,
-                                        properties=properties)
+
         publish = self._publish
         if retry:
             publish = self.connection.ensure(self, publish, **retry_policy)
-        return publish(message, routing_key, mandatory,
-                       immediate, exchange, declare)
+        return publish(body, priority, content_type,
+                       content_encoding, headers, properties,
+                        routing_key, mandatory, immediate, exchange, declare)
 
-    def _publish(self, message, routing_key, mandatory, immediate, exchange,
-            declare):
+    def _publish(self, body, priority, content_type, content_encoding,
+            headers, properties, routing_key, mandatory,
+            immediate, exchange, declare):
+        channel = self.channel
+        message = channel.prepare_message(
+            body, priority, content_type,
+            content_encoding, headers, properties,
+        )
         if declare:
             maybe_declare = self.maybe_declare
             [maybe_declare(entity) for entity in declare]
-        return self.exchange.publish(message, routing_key,
-                                     mandatory, immediate, exchange)
+        return channel.basic_publish(message,
+            exchange=exchange, routing_key=routing_key,
+            mandatory=mandatory, immediate=immediate,
+        )
+
+    def _get_channel(self):
+        channel = self._channel
+        if isinstance(channel, ChannelPromise):
+            channel = self._channel = channel()
+            self.exchange.revive(channel)
+            if self.on_return:
+                channel.events['basic_return'].add(self.on_return)
+        return channel
+
+    def _set_channel(self, channel):
+        self._channel = channel
+    channel = property(_get_channel, _set_channel)
 
     def revive(self, channel):
         """Revive the producer after connection loss."""
-        channel = self.channel = maybe_channel(channel)
-        self.exchange = self.exchange(channel)
-        self.exchange.revive(channel)
-
+        if is_connection(channel):
+            connection = channel
+            self.__connection__ = connection
+            channel = ChannelPromise(lambda: connection.default_channel)
+        if isinstance(channel, ChannelPromise):
+            self._channel = channel
+            self.exchange = self.exchange(channel)
+        else:
+            # Channel already concrete
+            self._channel = channel
+            if self.on_return:
+                self._channel.events['basic_return'].add(self.on_return)
+            self.exchange = self.exchange(channel)
         if self.auto_declare:
+            # auto_decare is not recommended as this will force
+            # evaluation of the channel.
             self.declare()
-        if self.on_return:
-            self.channel.events['basic_return'].append(self.on_return)
 
     def __enter__(self):
         return self
@@ -220,7 +252,7 @@ class Producer(object):
     @property
     def connection(self):
         try:
-            return self.channel.connection.client
+            return self.__connection__ or self.channel.connection.client
         except AttributeError:
             pass
 
@@ -233,13 +265,14 @@ class Consumer(object):
     :keyword no_ack: see :attr:`no_ack`.
     :keyword auto_declare: see :attr:`auto_declare`
     :keyword callbacks: see :attr:`callbacks`.
+    :keyword on_message: See :attr:`on_message`
     :keyword on_decode_error: see :attr:`on_decode_error`.
 
     """
     #: The connection/channel to use for this consumer.
     channel = None
 
-    #: A single :class:`~kombu.entity.Queue`, or a list of queues to
+    #: A single :class:`~kombu.Queue`, or a list of queues to
     #: consume from.
     queues = None
 
@@ -259,6 +292,25 @@ class Consumer(object):
     #: :class:`~kombu.transport.base.Message`).
     callbacks = None
 
+    #: Optional function called whenever a message is received.
+    #:
+    #: When defined this function will be called instead of the
+    #: :meth:`receive` method, and :attr:`callbacks` will be disabled.
+    #:
+    #: So this can be used as an alternative to :attr:`callbacks` when
+    #: you don't want the body to be automatically decoded.
+    #: Note that the message will still be decompressed if the message
+    #: has the ``compression`` header set.
+    #:
+    #: The signature of the callback must take a single argument,
+    #: which is the raw message object (a subclass of
+    #: :class:`~kombu.transport.base.Message`).
+    #:
+    #: Also note that the ``message.body`` attribute, which is the raw
+    #: contents of the message body, may in some cases be a read-only
+    #: :class:`buffer` object.
+    on_message = None
+
     #: Callback called when a message can't be decoded.
     #:
     #: The signature of the callback must take two arguments: `(message,
@@ -269,12 +321,13 @@ class Consumer(object):
     _next_tag = count(1).next   # global
 
     def __init__(self, channel, queues=None, no_ack=None, auto_declare=None,
-            callbacks=None, on_decode_error=None):
+            callbacks=None, on_decode_error=None, on_message=None):
         self.channel = channel
         self.queues = self.queues or [] if queues is None else queues
         self.no_ack = self.no_ack if no_ack is None else no_ack
         self.callbacks = (self.callbacks or [] if callbacks is None
                                                else callbacks)
+        self.on_message = on_message
         self._active_tags = {}
         if auto_declare is not None:
             self.auto_declare = auto_declare
@@ -454,7 +507,7 @@ class Consumer(object):
         """
         callbacks = self.callbacks
         if not callbacks:
-            raise NotImplementedError('Consumer does not have any callback')
+            raise NotImplementedError('Consumer does not have any callbacks')
         [callback(body, message) for callback in callbacks]
 
     def _basic_consume(self, queue, consumer_tag=None,
@@ -472,18 +525,18 @@ class Consumer(object):
         return tag
 
     def _receive_callback(self, message):
-        channel = self.channel
+        on_m, channel, decoded = self.on_message, self.channel, None
         try:
             m2p = getattr(channel, 'message_to_python', None)
             if m2p:
                 message = m2p(message)
-            decoded = message.decode()
+            decoded = None if on_m else message.decode()
         except Exception, exc:
             if not self.on_decode_error:
                 raise
             self.on_decode_error(message, exc)
         else:
-            self.receive(decoded, message)
+            return on_m(message) if on_m else self.receive(decoded, message)
 
     def __repr__(self):
         return '<Consumer: %s>' % (self.queues, )
