@@ -22,15 +22,15 @@ class Message(base.Message):
 class Channel(base.StdChannel):
     open = True
     throw_decode_error = False
-    _next_id = count(1).next
+    _ids = count(1)
 
     def __init__(self, connection):
         self.connection = connection
         self.called = []
         self.deliveries = count(1).next
         self.to_deliver = []
-        self.events = {'basic_return': []}
-        self.channel_id = self._next_id()
+        self.events = {'basic_return': set()}
+        self.channel_id = next(self._ids)
 
     def _called(self, name):
         self.called.append(name)
@@ -41,10 +41,10 @@ class Channel(base.StdChannel):
     def exchange_declare(self, *args, **kwargs):
         self._called('exchange_declare')
 
-    def prepare_message(self, message_data, properties={}, priority=0,
-            content_type=None, content_encoding=None, headers=None):
+    def prepare_message(self, body, priority=0, content_type=None,
+            content_encoding=None, headers=None, properties={}):
         self._called('prepare_message')
-        return dict(body=message_data,
+        return dict(body=body,
                     headers=headers,
                     properties=properties,
                     priority=priority,
@@ -93,6 +93,12 @@ class Channel(base.StdChannel):
     def basic_recover(self, requeue=False):
         self._called('basic_recover')
 
+    def exchange_bind(self, *args, **kwargs):
+        self._called('exchange_bind')
+
+    def exchange_unbind(self, *args, **kwargs):
+        self._called('exchange_unbind')
+
     def close(self):
         self._called('close')
 
