@@ -4,21 +4,25 @@ kombu.transport.pyamqp
 
 pure python amqp transport.
 
-:copyright: (c) 2009 - 2012 by Ask Solem.
-:license: BSD, see LICENSE for more details.
-
 """
 from __future__ import absolute_import
 
 import amqp
 
-from kombu.exceptions import StdChannelError
+from kombu.exceptions import (
+    StdConnectionError,
+    StdChannelError,
+    VersionMismatch,
+)
 from kombu.utils.amq_manager import get_manager
 
 from . import base
 
 DEFAULT_PORT = 5672
 
+if amqp.VERSION < (0, 9, 3):  # pragma: no cover
+    raise VersionMismatch('Please install amqp version 0.9.3 or higher.')
+
 
 class Message(base.Message):
 
@@ -38,11 +42,11 @@ class Message(base.Message):
 class Channel(amqp.Channel, base.StdChannel):
     Message = Message
 
-    def prepare_message(self, message_data, priority=None,
+    def prepare_message(self, body, priority=None,
                 content_type=None, content_encoding=None, headers=None,
                 properties=None):
         """Encapsulate data into a AMQP message."""
-        return amqp.Message(message_data, priority=priority,
+        return amqp.Message(body, priority=priority,
                             content_type=content_type,
                             content_encoding=content_encoding,
                             application_headers=headers,
@@ -64,7 +68,8 @@ class Transport(base.Transport):
 
     # it's very annoying that pyamqp sometimes raises AttributeError
     # if the connection is lost, but nothing we can do about that here.
-    connection_errors = amqp.Connection.connection_errors
+    connection_errors = (StdConnectionError, ) + \
+                        amqp.Connection.connection_errors
     channel_errors = (StdChannelError, ) + amqp.Connection.channel_errors
 
     nb_keep_draining = True
