@@ -6,9 +6,6 @@ Execute cleanup handlers when objects go out of scope.
 
 Taken from :class:`multiprocessing.util.Finalize`.
 
-:copyright: (c) 2009 - 2012 by Ask Solem.
-:license: BSD, see LICENSE for more details.
-
 """
 from __future__ import absolute_import
 
@@ -16,7 +13,7 @@ import weakref
 
 from itertools import count
 
-__all__ = ["Finalize"]
+__all__ = ['Finalize']
 
 
 class Finalize(object):
@@ -72,7 +69,7 @@ class Finalize(object):
         try:
             obj = self._weakref()
         except (AttributeError, TypeError):
-            return "<Finalize: (dead)>"
+            return '<Finalize: (dead)>'
 
         if obj is None:
             return
